'''
Session module
(part of web.py)
'''


import time
import random

try:				# Python 2.5
	from hashlib import sha512 as sha
except ImportError:
	import sha

try:				# faster C-implementation
	import cPickle as pickle
except ImportError:
	import pickle

import web


session_parameters = web.utils.storage({
	'cookie_name': 'webpy',		# name of the cookie which will transfer the session id
	'cookie_domain': None,		# cookie domain for the setcookie() when setting session id cookie
	'timeout': 600,
	'max_age': 24 * 60 * 60,
	'id_seed': 'web.py',
	'regenerate_id': True,		# boolean, on every request regenerate id and set again cookie
	'generator': False,			# if False, use default generator

	'ignore_expiration': False,
	'ignore_old_age': False,
	'ignore_change_ip': True,	# boolean, if the pair ( id, ip ) doesn't match the db, then fail/raise exception/...

	'handler': 'db'
})

handler_parameters = web.utils.storage({
	'file_dir': '/tmp',			# storage dir
	'db_table': 'session_data'	# table name
})


class Session(web.utils.Storage):
	'''main session object
		main instance variables
		_generator - reference to user-supplied "unique random number generator" or just to default session generator function
		_handler
		_id
		_data - internal Storage object'''

	def __init__(self):
		self._id = False
		self._old_id = False
		self._data = {}

	def __getattr__(self, key):
		if key in ('_id', '_old_id', '_data', '_handler', '_generator'):
			return object.__getattr__(self, key)

		try:
			return self._data[key]
		except KeyError, k:
			raise AttributeError, k

	def __setattr__(self, key, value):
		if key in ('_id', '_old_id', '_data', '_handler', '_generator'):
			object.__setattr__(self, key, value)
		else:
			self._data[key] = value

	def __delattr__(self, key):
		if key in ('_id', '_old_id', '_data', '_handler', '_generator'):
			#object.__delattr__(self, key, value)
			return

		try:
			del self._data[key]
		except KeyError, k:
			raise AttributeError, k

	def __getitem__(self, key):
		return self._data[key]

	def __setitem__(self, key, value):
		self._data[key] = value

	def __delitem__(self, key):
		del self._data[key]

	def __repr__(self):
		return '<Session {\'id\': ' + str(self._id) + ', \'data\': ' + dict.__repr__(self._data) + '}>'

	# public methods
	def start(self):
		'''starts the session: creates handler object, sets generator,
		regenerates id, sets cookies, calls _identity, _verify, _generator function'''
		if not web.config.session_parameters.generator:
			self._generator = self._generate_id
		else:
			self._generator = web.config.session_parameters.generator

		handler = {'db': DBHandler, 'file': FileHandler}
		self._handler = handler[web.config.session_parameters.handler]()

		self._identify()

		if self._id:
			item = self._retreive()
			if self._verify(item):
				self._data = item.data

		if web.config.session_parameters.regenerate_id:
			self._old_id = self._id
			self._id = self._generator()

		if not self._id:
			self._id = self._generator()

		web.setcookie(web.config.session_parameters.cookie_name, self._id, web.config.session_parameters.timeout, web.config.session_parameters.cookie_domain)

	def get_id(self):
		'''returns current session id'''
		return self._id

	def cleanup(self):
		'''cleans expired sessions'''
		self._handler.clean(web.config.session_parameters.timeout)

	def save(self):
		'''save session data'''
		self._store()

	def destroy(self):
		'''removes session (including cookies)'''
		self._remove()
		self._id = False
		self._old_id = False
		web.setcookie(web.config.session_parameters.cookie_name, '', (-1) * web.config.session_parameters.timeout, web.config.session_parameters.cookie_domain)

	# private methods
	def _generate_id(self):
		'''implicit session id generator
		"hashes" ip, time, seed ...'''
		seed = '%s %s %s %s' % (random.random(), time.time(), web.ctx.ip, web.config.session_parameters.id_seed)

		try:		# >= Python 2.5
			return sha(seed).hexdigest()
		except TypeError:
			return sha.new(seed).hexdigest()

	def _identify(self):
		'''identifies session id (through cookies)'''
		try:
			self._id = web.cookies().__getattr__(web.config.session_parameters.cookie_name)
		except AttributeError, KeyError:
			pass

	def _verify(self, item):
		'''verifies with retreived data from handler object'''
		try:
			old_ip = item.ip
		except AttributeError:
			self._id = False
			return False

		if not web.config.session_parameters.ignore_expiration	\
			and time.time() - item.time >= web.config.session_parameters.timeout:
			self._remove()
			self._id = False
			return False

		if not web.config.session_parameters.ignore_change_ip	\
			and web.ctx.ip != item.ip.strip():
			self._remove()
			self._id = False
			return False

		if not web.config.session_parameters.ignore_old_age	\
			and time.time() - item.created >= web.config.session_parameters.max_age:
			self._remove()
			self._id = False
			return False

		return True

	def _store(self):
		'''stores session data (wrapper around handler object)'''
		self._handler.store(self._id, web.ctx.ip, self._data, self._old_id)

	def _retreive(self):
		'''retreive session data (wrapper around handler object)'''
		return self._handler.retreive(self._id)

	def _remove(self):
		'''removes session data (wrapper around handler object)'''
		self._handler.remove(self._id)

		if self._old_id:
			self._handler.remove(self._old_id)


class Handler:
	'''abstract handler class'''
	def __init__(self):
		pass

	def store(self, id, ip, data, old_id=False):
		'''takes
			ip - client ip
			id - string
			data - dictionary/storage
			old_id - if the id regenerates after every request'''
		pass

	def retreive(self, id):
		'''returns storage'''
		pass

	def remove(self, id):
		pass

	def clean(self, timeout):
		'''removes all expired sessions'''
		pass


class DBHandler(Handler):
	'''needs an table:
	CREATE TABLE session_data (
		id CHAR(129) UNIQUE NOT NULL,
		ip CHAR(16) NOT NULL,
		created int NOT NULL,
		touched int NOT NULL,
		data TEXT
	);'''
	def __init__(self):
		pass

	def store(self, id, ip, data, old_id=False):
		'''takes
			ip - client ip
			id - string
			data - dictionary/storage
			old_id - if the id regenerates after every request'''
		do_insert = True

		if not old_id:
			old_id = id

		if len(list(web.select(web.config.handler_parameters.db_table,
				vars={'id': old_id},
				what='id',
				where='id = $id'))) == 1:
			do_insert = False

		web.transact()

		t = int(time.time())
		try:
			if do_insert:
				web.db.insert(web.config.handler_parameters.db_table, seqname=False,
					id=id, ip=ip, touched=t, created=t,
					data=pickle.dumps(data, 0))
			else:
				web.update(web.config.handler_parameters.db_table,
					where='id = $old_id',
					vars={'old_id': old_id},
					id=id, ip=ip, touched=t,
					data=pickle.dumps(data, 0))
			web.commit()

		except Exception, inst:
			web.rollback()
			raise inst

	def remove(self, id):
		web.transact()
		try:
			web.delete(web.config.handler_parameters.db_table,
					where='id = $id', vars={'id': id})
			web.commit()

		except Exception, inst:
			web.rollback()
			raise inst

	def retreive(self, id):
		'''returns storage'''
		try:
			iter = web.select(web.config.handler_parameters.db_table,
					what='*',
					vars={
							'id': id,
							'timeout': web.config.session_parameters.timeout
						},
					where='id = $id')

		except Exception, inst:
			raise inst

		try:
			result = iter[0]

			result.data = pickle.loads(result.data.encode('ascii'))
			result.time = result.touched
			del result.touched

			return result

		except IndexError:
			return web.Storage()

	def clean(self, timeout):
		'''removes all expired sessions'''
		web.transact()

		try:
			web.delete(web.config.handler_parameters.db_table,
				where='($now - touched) >= $timeout',
				vars={'timeout': timeout, 'now': int(time.time())})
			web.commit()

		except Exception, inst:
			web.rollback()
			raise inst


def FileHandler(Handler):
	def __init__(self):
		'''takes dir where to create session files'''
		pass

	def store(self, id, ip, data, old_id=False):
		'''takes
			ip - client ip
			id - string
			data - dictionary/storage
			old_id - if the id regenerates after every request'''
		pass

	def retreive(self, id):
		'''returns storage'''
		pass

	def remove(self, id):
		pass


def _load_session():
	web.webapi.config.session_parameters = session_parameters
	web.webapi.config.handler_parameters = handler_parameters
	web.webapi.ctx.session = Session()

def _unload_session():
	if web.ctx.session.get_id():
		web.ctx.session.save()


web.webapi.loadhooks['session'] = _load_session
web.webapi.unloadhooks['session'] = _unload_session

